#include <Malibu_Mb.h>
#include <Analysis_Mb.h>
#include <Exception_Mb.h>
#include <SystemSupport_Mb.h>
#include <StringSupport_Mb.h>
#include <sstream>
#include <limits>
#include <cmath>
#include <numeric>
#include <Random_Mb.h>
#include <Magic_Mb.h>
#include <FileSupport_Mb.h>
#include <DeviceEnum_Mb.h>
#include <IppSupport_Mb.h>
#include <Buffer_Mb.h>
#include <BufferHeader_Mb.h>
#include <BufferDatagrams_Mb.h>
#include <IppMemoryUtils_Mb.h>
#include <PacketStream_Mb.h>
#include <CalMemBase_Mb.h>
#include <X3_SD16_Mb.h>
#include <StaticWaveform_App.h>

using namespace Innovative;
using namespace InnovativeKernel;
using namespace std;

X3_SD16 *m_X3SD16;
PacketStream m_X3SD16Stream;
Innovative::Buffer m_X3SD16WaveformPacketTX;

#define ERROR printf
#define DEBUG printf

enum T_DAC_NAME {
    E_DAC_LF_I = 1,
    E_DAC_LF_Q = 2,
    E_DAC_LF_G = 5,
    E_DAC_LF_P = 9,

    E_DAC_HF_I = 3,
    E_DAC_HF_Q = 4,
    E_DAC_HF_G = 6,
    E_DAC_HF_P = 10,

    E_DAC_AUDIO_HEADPHONE = 7,
    E_DAC_AUDIO_SPK = 8,
};

enum T_ADC_NAME {
    E_ADC_LF_I = 1,
    E_ADC_LF_Q = 2,
    E_ADC_LF_P = 5,

    E_ADC_HF_I = 3,
    E_ADC_HF_Q = 4,
    E_ADC_HF_P = 6,

    E_ADC_IMM = 7,
};

const int K_RXDataSize = 2048;
const int K_TXDataSize = (K_RXDataSize << 1);
const double K_SamplingFrequency = 12286.376953; /* X3-SD16 sampling frequency in Hz */

class CX3SD16
{
public:
    int Init();
    bool ReadRom();
    void Close();
private:
    X3_SD16 *m_X3SD16;


    /* ----------------------------------- */
    /* -------- HANDLE LOGIC DOWNLOAD ---- */
    /* ----------------------------------- */
    void HandleFileReadProgress(Innovative::ProcessProgressEvent & event);
    void HandleFileReadComplete(Innovative::ProcessCompletionEvent & event);
    void HandleParseProgress(Innovative::ProcessProgressEvent & event);
    void HandleParseComplete(Innovative::ProcessCompletionEvent & event);
    void HandleParseMessage(Innovative::ProcessStatusEvent & event);
    void HandleLoadProgress(Innovative::ProcessProgressEvent & event);
    void HandleLoadComplete(Innovative::ProcessCompletionEvent & event);
    void HandleLoadMessage(Innovative::ProcessStatusEvent & event);

    /* ----------------------------------- */
    /* -------- HANDLE ALERT TEMPERATURE - */
    /* ----------------------------------- */
    void HandleWarningTempAlert(Innovative::AlertSignalEvent & event);

    /* ----------------------------------- */
    /* -------- HANDLE ALERT INPUT ------ */
    /* ----------------------------------- */
    void HandleInputFifoOverrangeAlert(Innovative::AlertSignalEvent & event);

    /* ----------------------------------- */
    /* -------- HANDLE ALERT OUTPUT ------ */
    /* ----------------------------------- */
    void HandleOutputFifoUnderrunAlert(Innovative::AlertSignalEvent & event);
    void HandleOutputTriggerAlert(Innovative::AlertSignalEvent & event);
    void HandlePllLostAlert(Innovative::AlertSignalEvent & event);

    /* ----------------------------------- */
    /* -------- HANDLE STREAM ------------ */
    /* ----------------------------------- */
    void HandleBeforeStreamStart(OpenWire::NotifyEvent & event);
    void HandleAfterStreamStart(OpenWire::NotifyEvent & event);
    void HandleAfterStreamStop(OpenWire::NotifyEvent & event);

    /* ----------------------------------- */
    /* -------- HANDLE DATA -------------- */
    /* ----------------------------------- */
    /*
    void HandleAfterStart(OpenWire::NotifyEvent & event);
    void HandleAfterStop(Innovative::NotifyEvent & event);
    void HandleBeforeStart(Innovative::NotifyEvent & event);
    void HandleDirectDataAvailable(Innovative::PacketStreamDirectDataEvent & Event);
    void HandleDirectDataRequired(Innovative::PacketStreamDirectDataEvent & Event);
    void HandleDataRequired(Innovative::PacketStreamDataEvent & Event);
     */
    void HandleDataAvailable(Innovative::PacketStreamDataEvent & Event);

    volatile bool m_FPGADownloaded;
    volatile bool m_PLLLost;
};

void CX3SD16::HandleFileReadProgress(Innovative::ProcessProgressEvent & event) {
    printf("HandleFileReadProgress %d\n", event.Percent);
}

void CX3SD16::HandleFileReadComplete(Innovative::ProcessCompletionEvent & event) {
    printf("HandleFileReadComplete %d\n", event.Result);
}

void CX3SD16::HandleParseProgress(Innovative::ProcessProgressEvent & event) {
    printf("HandleParseProgress %d\n", event.Percent);
}

void CX3SD16::HandleParseComplete(Innovative::ProcessCompletionEvent & event) {
    printf( "HandleParseComplete %d\n", event.Result);
}

void CX3SD16::HandleParseMessage(Innovative::ProcessStatusEvent & event) {
    printf( "HandleParseMessage\n", 0);
}

void CX3SD16::HandleLoadProgress(Innovative::ProcessProgressEvent & event) {
    printf("HandleLoadProgress %d\n", event.Percent);
}

void CX3SD16::HandleLoadComplete(Innovative::ProcessCompletionEvent & event) {
    printf( "HandleLoadComplete %d\n", event.Result);
    m_FPGADownloaded = true;
}

void CX3SD16::HandleLoadMessage(Innovative::ProcessStatusEvent & event) {
    printf("HandleLoadMessage %s\n\n", event.Message.c_str());
}

void CX3SD16::HandleWarningTempAlert(Innovative::AlertSignalEvent & event) {
    printf("HandleWarningTempAlert\n", 0);
}

void CX3SD16::HandleBeforeStreamStart(OpenWire::NotifyEvent & event) {
    printf("HandleBeforeStreamStart\n", 0);
}

void CX3SD16::HandleAfterStreamStart(OpenWire::NotifyEvent & event) {
    printf("HandleAfterStreamStart\n", 0);
}

void CX3SD16::HandleAfterStreamStop(OpenWire::NotifyEvent & event) {
    printf("HandleAfterStreamStop\n", 0);
}

void CX3SD16::HandleOutputFifoUnderrunAlert(Innovative::AlertSignalEvent & event) {
    printf("HandleOutputFifoUnderrunAlert\n", 0);
}

void CX3SD16::HandleOutputTriggerAlert(Innovative::AlertSignalEvent & event) {
    printf("HandleOutputTriggerAlert\n", 0);
}

void CX3SD16::HandlePllLostAlert(Innovative::AlertSignalEvent & event) {
    m_PLLLost = true;
    printf("HandlePllLostAlert\n", 0);
}

/* ----------------------------------- */
/* -------- HANDLE INPUT ------------- */

/* ----------------------------------- */

void CX3SD16::HandleInputFifoOverrangeAlert(Innovative::AlertSignalEvent & event) {
    printf("HandleInputFifoOverrangeAlert\n", 0);
}

/* ----------------------------------- */
/* -------- HANDLE DATA -------------- */
/* ----------------------------------- */

/*
void  CX3SD16::HandleAfterStart(Innovative::NotifyEvent & event)
{
    printf("HandleAfterStart\n");
}

void  CX3SD16::HandleAfterStop(Innovative::NotifyEvent & event)
{
    printf("HandleAfterStop\n");
}

void  CX3SD16::HandleBeforeStart(Innovative::NotifyEvent & event)
{
    printf("HandleBeforeStart\n");
}

void CX3SD16::HandleDirectDataRequired(Innovative::PacketStreamDirectDataEvent & Event)
{
    printf("HandleDirectDataRequired\n", 0);
}

void CX3SD16::HandleDirectDataAvailable(Innovative::PacketStreamDirectDataEvent & Event)
{
    printf("HandleDirectDataAvailable\n", 0);
}

void CX3SD16::HandleDataRequired(Innovative::PacketStreamDataEvent & Event)
{

}
 */


bool CX3SD16::ReadRom() {
    bool ret;

    ret = true;
    m_X3SD16->IdRom().LoadFromRom();

    printf("IdRom().Name():%s\n", m_X3SD16->IdRom().Name().c_str());
    printf("IdRom().Revision():%s\n", m_X3SD16->IdRom().Revision().c_str());

    for (unsigned int range = 0; range < 3; ++range) {
        for (unsigned int ch = 0; ch < 16; ++ch) {
            printf("Range:%02d Channel:%02d Gain:%f Offset:%f\n", range, ch, m_X3SD16->Input().Gain(range, ch), m_X3SD16->Input().Offset(range, ch));
        }

        printf("Range:%d Calibrated:%d\n", range, m_X3SD16->Input().Calibrated(range));
        if (m_X3SD16->Input().Calibrated(range) == 0) {
            printf("Strange, range %d is not calibrated :(\n", range);
            ret = false;
        }
    }

    for (unsigned int ch = 0; ch < 16; ++ch) {
        printf("Output Channel:%02d Gain:%f Offset:%f\n", ch, m_X3SD16->Output().Gain(ch), m_X3SD16->Output().Offset(ch));
    }

    printf("Output Calibrated:%d\n", m_X3SD16->Output().Calibrated());
    if (m_X3SD16->Output().Calibrated() == 0) {
        printf("Strange, Output is not calibrated :(\n", 0);
        ret = false;
    }

    return ret;
}

int m_NbChannelsRX;
int m_NbChannelsTX;


void CX3SD16::HandleDataAvailable(Innovative::PacketStreamDataEvent & Event) {
    Buffer Packet;
    int DataRx1[K_RXDataSize];
    int DataRx2[K_RXDataSize];

    int *PtData = (int *) m_X3SD16WaveformPacketTX.Data();

    for (unsigned int i = 0; i < K_TXDataSize; i++) {
	for (unsigned int j = 0; j < m_NbChannelsTX; j++) {
	   *(PtData + i * m_NbChannelsTX + j) = 0;
	}
    }

    m_X3SD16Stream.Send(m_X3SD16WaveformPacketTX);

    // Data (Signal, TxPulse) reception
    Event.Sender->Recv(Packet);
    const int *pBuffer = Packet.ConstData();

    // Get Data from ch0 & ch1
    for (int i = 0; i < K_RXDataSize; i++) {
        DataRx1[i] = *(pBuffer+E_ADC_HF_I-1);
        DataRx2[i] = *(pBuffer+E_ADC_HF_Q-1);
        pBuffer += m_NbChannelsRX;
    }

    // Monitor Means Vals
    double mean1, mean2;
    int max1, max2;

    mean1 = 0;
    mean2 = 0;
    max1 = 0;
    max2 = 0;

/*
   FILE *fp = fopen("/tmp/trace.txt", "w");


    for (int i = 0; i < K_RXDataSize; i++) {
	mean1 += fabs(DataRx1[i]);
	mean2 += fabs(DataRx2[i]);
        if (DataRx1[i]>max1)
           max1 = DataRx1[i];
        if (DataRx2[i]>max2)
           max2 = DataRx2[i];

        fprintf(fp, "%d;%d\n", DataRx1[i], DataRx2[i]);
    }

    fclose(fp);
*/
    mean1 /= K_RXDataSize;
    mean2 /= K_RXDataSize;


    printf("%f %f %d %d\n", mean1, mean2, max1, max2);
}

void CX3SD16::Close()
{
    m_X3SD16Stream.Stop();
    m_X3SD16Stream.Disconnect();
    m_X3SD16->Close();
}


int CX3SD16::Init()
{
    try {
        m_X3SD16 = new X3_SD16;
    } catch (Innovative::MalibuException & e) {
        printf("m_X3SD16 Device Creation Failure : %s\n\n", e.what());
        m_X3SD16 = NULL;
        return -1;
    }

    // Use IPP performance memory functions.
    Init::UsePerformanceMemoryFunctions();

    //  Configure Event Handlers
    m_X3SD16->Logic().OnFpgaFileReadProgress.SetEvent(this, &CX3SD16::HandleFileReadProgress);
    m_X3SD16->Logic().OnFpgaFileReadComplete.SetEvent(this, &CX3SD16::HandleFileReadComplete);
    m_X3SD16->Logic().OnFpgaParseProgress.SetEvent(this, &CX3SD16::HandleParseProgress);
    m_X3SD16->Logic().OnFpgaParseComplete.SetEvent(this, &CX3SD16::HandleParseComplete);
    m_X3SD16->Logic().OnFpgaParseMessage.SetEvent(this, &CX3SD16::HandleParseMessage);
    m_X3SD16->Logic().OnFpgaLoadProgress.SetEvent(this, &CX3SD16::HandleLoadProgress);
    m_X3SD16->Logic().OnFpgaLoadComplete.SetEvent(this, &CX3SD16::HandleLoadComplete);
    m_X3SD16->Logic().OnFpgaLoadMessage.SetEvent(this, &CX3SD16::HandleLoadMessage);

    // Configure Stream
    m_X3SD16->OnBeforeStreamStart.SetEvent(this, &CX3SD16::HandleBeforeStreamStart);
    m_X3SD16->OnBeforeStreamStart.Synchronize();
    m_X3SD16->OnAfterStreamStart.SetEvent(this, &CX3SD16::HandleAfterStreamStart);
    m_X3SD16->OnAfterStreamStart.Synchronize();
    m_X3SD16->OnAfterStreamStop.SetEvent(this, &CX3SD16::HandleAfterStreamStop);
    m_X3SD16->OnAfterStreamStop.Synchronize();
    //    m_X3SD16Stream.OnAfterStart(this, &CX3SD16::HandleAfterStart);
    //    m_X3SD16Stream.OnAfterStop(this, &CX3SD16::HandleAfterStop);
    //    m_X3SD16Stream.OnBeforeStart(this, &CX3SD16::HandleBeforeStart);
    //    m_X3SD16Stream.OnDirectDataAvailable.SetEvent(this, &CX3SD16::HandleDirectDataAvailable);
    //    m_X3SD16Stream.OnDirectDataRequired.SetEvent(this, &CX3SD16::HandleDirectDataRequired);
    //    m_X3SD16Stream.OnDataRequired.SetEvent(this, &CX3SD16::HandleDataRequired);
    m_X3SD16Stream.OnDataAvailable.SetEvent(this, &CX3SD16::HandleDataAvailable);
    m_X3SD16Stream.DirectDataMode(false);

    //  Configure Alerts
    //    m_X3SD16->Alerts().OnTimeStampRolloverAlert.SetEvent(this, &ApplicationIo::HandleTimestampRolloverAlert);
    //    m_X3SD16->Alerts().OnSoftwareAlert.SetEvent(this, &ApplicationIo::HandleSoftwareAlert);
    //    m_X3SD16->Alerts().OnWarningTemperature.SetEvent(this, &CX3SD16::HandleWarningTempAlert);
    //    m_X3SD16->Alerts().AlertEnable(IUsesX3Alerts::alertWarningTemperature, true);
    m_X3SD16->Alerts().OnOutputFifoUnderrun.SetEvent(this, &CX3SD16::HandleOutputFifoUnderrunAlert);
    m_X3SD16->Alerts().AlertEnable(IUsesX3Alerts::alertOutputFifoUnderrun, true);
    m_X3SD16->Alerts().OnOutputTrigger.SetEvent(this, &CX3SD16::HandleOutputTriggerAlert);
    m_X3SD16->Alerts().AlertEnable(IUsesX3Alerts::alertOutputTrigger, true);
    m_X3SD16->Alerts().OnPllLost.SetEvent(this, &CX3SD16::HandlePllLostAlert);
    m_X3SD16->Alerts().AlertEnable(IUsesX3Alerts::alertPllLost, true);
    //    m_X3SD16().Alerts().OnPllLost.SetEvent(this, &ApplicationIo::HandlePllLostAlert);
    //    m_X3SD16().Alerts().OnInputFifoOverrun.SetEvent(this, &ApplicationIo::HandleInputFifoOverrunAlert);
    //    m_X3SD16().Alerts().OnInputTrigger.SetEvent(this, &ApplicationIo::HandleInputTriggerAlert);
    m_X3SD16->Alerts().OnInputOverrange.SetEvent(this, &CX3SD16::HandleInputFifoOverrangeAlert);
    m_X3SD16->Alerts().AlertEnable(IUsesX3Alerts::alertInputOverrange, true);
    //    m_X3SD16().Alerts().OnInputStopTrigger.SetEvent(this, &ApplicationIo::HandleInputStopTriggerAlert);
    //    m_X3SD16().Alerts().OnInputPretriggerError.SetEvent(this, &ApplicationIo::HandleInputPretriggerErrorAlert);

    //  Open Devices
    const int Meg = 1024 * 1024;
    int FBusmasterSize = (4) * Meg;
    m_X3SD16->IncomingBusMasterSize(FBusmasterSize);
    m_X3SD16->OutgoingBusMasterSize(FBusmasterSize);

    m_X3SD16->Target(0);

    try {
        m_X3SD16->Open();
    } catch (Innovative::MalibuException & e) {
        printf("m_X3SD16 Device Open Failure : %s\n\n", e.what());
        m_X3SD16 = NULL;
        return -2;
    }

    m_FPGADownloaded = false;
    m_X3SD16->Logic().ConfigureFpga("fpga_nuss_2f_mk2.bit");

    while (m_FPGADownloaded == false) {
        printf("Waiting for FPGA to be downloaded...\n", 0);
        sleep(1);
    }

    while (m_X3SD16->Logic().IsConfigured() != 1) {
        printf("Waiting for Logic to be configured...\n", 0);
        sleep(1);
    }

    printf("Logic is configured...\n", 0);
    
    m_X3SD16->Reset();

    ReadRom();

    //  Connect Stream
    if (m_X3SD16Stream.ConnectTo(m_X3SD16) == true)
        printf("Stream Connected\n", 0);
    else
        printf("Stream Not Connected\n", 0);

    printf("\t \n", 0);
    printf("- m_X3SD16->Thermal().LogicFailureTemperature() : %f\n", m_X3SD16->Thermal().LogicFailureTemperature());
    printf("- m_X3SD16->Thermal().LogicWarningTemperature() : %f\n", m_X3SD16->Thermal().LogicWarningTemperature());
    printf("- m_X3SD16->Thermal().LogicTemperature() : %f\n", m_X3SD16->Thermal().LogicTemperature());
    printf("- m_X3SD16->Thermal().Failed() : %d\n", m_X3SD16->Thermal().Failed());
    printf("\t \n", 0);
    printf("- m_X3SD16->PciLogicVersion().PciLogicFamily() = %04x\n", m_X3SD16->PciLogicVersion().PciLogicFamily());
    printf("- m_X3SD16->PciLogicVersion().PciLogicPcb() = %04x\n", m_X3SD16->PciLogicVersion().PciLogicPcb());
    printf("- m_X3SD16->PciLogicVersion().PciLogicRawValue() = %04x\n", m_X3SD16->PciLogicVersion().PciLogicRawValue());
    printf("- m_X3SD16->PciLogicVersion().PciLogicRevision() = %04x\n", m_X3SD16->PciLogicVersion().PciLogicRevision());
    printf("- m_X3SD16->PciLogicVersion().PciLogicType() = %04x\n", m_X3SD16->PciLogicVersion().PciLogicType());
    printf("\t \n", 0);
    printf("- m_X3SD16->UserLogicVersion().FpgaLogicRawValue() = %04x\n", m_X3SD16->UserLogicVersion().FpgaLogicRawValue());
    printf("- m_X3SD16->UserLogicVersion().FpgaLogicRevision() = %04x\n", m_X3SD16->UserLogicVersion().FpgaLogicRevision());
    printf("- m_X3SD16->UserLogicVersion().FpgaLogicType() = %04x\n", m_X3SD16->UserLogicVersion().FpgaLogicType());
    printf("- m_X3SD16->UserLogicVersion().FpgaLogicVersion() = %04x\n", m_X3SD16->UserLogicVersion().FpgaLogicVersion());
    printf("- m_X3SD16->UserLogicVersion().FpgaLogicVariant() = %04x\n", m_X3SD16->UserLogicVersion().FpgaLogicVariant());
    printf("\t \n", 0);
    printf("- m_X3SD16->Input().Info().MaxRate() = %f\n", m_X3SD16->Input().Info().MaxRate());
    printf("- m_X3SD16->Input().Info().MinRate() = %f\n", m_X3SD16->Input().Info().MinRate());
    printf("- m_X3SD16->Input().Info().BufferSize() = %d\n", m_X3SD16->Input().Info().BufferSize());
    printf("- m_X3SD16->Input().Info().Channels().Channels() = %d\n", m_X3SD16->Input().Info().Channels().Channels());
    printf("- m_X3SD16->Input().Info().ClockFactor() = %d\n", m_X3SD16->Input().Info().ClockFactor());
    printf("- m_X3SD16->Input().Info().GangSize() = %d\n", m_X3SD16->Input().Info().GangSize());
    printf("- m_X3SD16->Input().Info().ReferenceClock() = %f\n", m_X3SD16->Input().Info().ReferenceClock());
    printf("\t \n", 0);

    // We use internal reference source
    m_X3SD16->ReferenceSource(X3_SD16::rsCrystal);
    m_X3SD16->Clock().Reference(m_X3SD16->Input().Info().ReferenceClock());
    m_X3SD16->Clock().OutputClock(X3_SD16::Timebase::oVco);
    m_X3SD16->Clock().ReferenceCalibrationFactor(1.0);
    m_X3SD16->Clock().Enabled(true);
    m_X3SD16->Clock().Frequency(K_SamplingFrequency);
    printf("- m_X3SD16->Clock().FrequencyActual() = %f\n", m_X3SD16->Clock().FrequencyActual());

    // Disable triggering initially
    m_X3SD16->Input().ExternalTrigger(false);
    m_X3SD16->Output().ExternalTrigger(false);
    m_X3SD16->Trigger(X3_SD16::tOutput, false);
    m_X3SD16->Trigger(X3_SD16::tInput, false);

    // Set sync source -> board is Master
    m_X3SD16->Master(true);
    
    printf("- m_X3SD16->Clock().Index() = %d\n", m_X3SD16->Clock().Index());
    printf("- m_X3SD16->Clock().ClockFactor() = %f\n", m_X3SD16->Clock().ClockFactor());
    printf("- m_X3SD16->Clock().Enabled() = %d\n", m_X3SD16->Clock().Enabled());
    printf("- m_X3SD16->Clock().FrequencyActual() = %f\n", m_X3SD16->Clock().FrequencyActual());
    printf("- m_X3SD16->Clock().Frequency() = %f\n", m_X3SD16->Clock().Frequency());
    printf("- m_X3SD16->Clock().Period() = %f\n", m_X3SD16->Clock().Period());
    printf("- m_X3SD16->Clock().Reference() = %f\n", m_X3SD16->Clock().Reference());
    printf("- m_X3SD16->Clock().ReferenceCalibrationFactor() = %f\n", m_X3SD16->Clock().ReferenceCalibrationFactor());
    printf("\t \n", 0);

    // Acq and Playback are continous.
    m_X3SD16->Input().Unframed();
    m_X3SD16->Output().Unframed();

    m_X3SD16Stream.PrefillPacketCount(10);
    printf("Stream : PrefillPacketCount()=%d\n", m_X3SD16Stream.PrefillPacketCount());

    // OUTPUT Channels Configuration
    // -----------------------------
    m_X3SD16->Output().Info().Channels().DisableAll();

    // LF
    m_X3SD16->Output().Info().Channels().Enabled(E_DAC_LF_I, true); // Re
    m_X3SD16->Output().Info().Channels().Enabled(E_DAC_LF_Q, true); // Im
    m_X3SD16->Output().Info().Channels().Enabled(E_DAC_LF_P, true); // TX Pulse
    m_X3SD16->Output().Info().Channels().Enabled(E_DAC_LF_G, true); // Gain RX

    // HF
    m_X3SD16->Output().Info().Channels().Enabled(E_DAC_HF_I, true); // Re
    m_X3SD16->Output().Info().Channels().Enabled(E_DAC_HF_Q, true); // Im
    m_X3SD16->Output().Info().Channels().Enabled(E_DAC_HF_P, true); // TX Pulse
    m_X3SD16->Output().Info().Channels().Enabled(E_DAC_HF_G, true); // Gain RX

    // Audio
    m_X3SD16->Output().Info().Channels().Enabled(E_DAC_AUDIO_HEADPHONE, true); // Audio Line
    m_X3SD16->Output().Info().Channels().Enabled(E_DAC_AUDIO_SPK, true); // Audio Speaker

    // INPUT Channels Configuration
    // ----------------------------
    m_X3SD16->Input().Info().Channels().DisableAll();

    for (int i = 0; i < 16; i++)
        m_X3SD16->Input().Range(i, X3RangedIoDevice::x1);

    // LF
    m_X3SD16->Input().Info().Channels().Enabled(E_ADC_LF_I, true); // Re
    m_X3SD16->Input().Info().Channels().Enabled(E_ADC_LF_Q, true); // Im
    m_X3SD16->Input().Info().Channels().Enabled(E_ADC_LF_P, true); // TxPulse

    // HF
    m_X3SD16->Input().Info().Channels().Enabled(E_ADC_HF_I, true); // Re
    m_X3SD16->Input().Info().Channels().Enabled(E_ADC_HF_Q, true); // Im
    m_X3SD16->Input().Info().Channels().Enabled(E_ADC_HF_P, true); // TxPulse

    // IMM
    // Immersion channel will be configured for SNLE program only
    m_X3SD16->Input().Info().Channels().Enabled(E_ADC_IMM, true);

    m_NbChannelsRX = m_X3SD16->Input().Info().Channels().ActiveChannels();
    m_NbChannelsTX = m_X3SD16->Output().Info().Channels().ActiveChannels();

    int PacketSizeRX = K_RXDataSize * m_NbChannelsRX + 2;
    int PacketSizeTX = K_TXDataSize * m_NbChannelsTX + 2;

    // Configure outputs
    m_X3SD16->Output().PacketSize(PacketSizeTX);
    m_X3SD16->Output().Decimation(1);
    m_X3SD16->Output().PretriggerCount(0);

    // Configure inputs
    m_X3SD16->Input().PacketSize(PacketSizeRX);
    m_X3SD16->Input().Decimation(1);
    m_X3SD16->Input().PretriggerCount(0);

    // Calculate waveform buffer
    IntegerDG Packet_DG(m_X3SD16WaveformPacketTX);
    //
    // Calculate Packet Size in int
    Packet_DG.Resize(K_TXDataSize * m_NbChannelsTX);

    PacketBufferHeader PktBufferHdr(m_X3SD16WaveformPacketTX);
    PktBufferHdr.PacketSize(K_TXDataSize * m_NbChannelsTX);
    PktBufferHdr.PeripheralId(m_X3SD16->Output().PacketId());
    PktBufferHdr[1] = 0x00000000;
    int *PtData = (int *) m_X3SD16WaveformPacketTX.Data();
    for (int i = 0; i < K_TXDataSize * m_NbChannelsTX; i++)
        PtData[i] = 0;

    // PortConfig 0xff=Output.
    m_X3SD16->FrontPanel().FrontPanelPortConfig(0xff);
    m_X3SD16->FrontPanel().FrontPanelPortData().Value(0x0);

    m_X3SD16Stream.Start();

    for (int i = 0; i < 2; i++)
        m_X3SD16Stream.Send(m_X3SD16WaveformPacketTX);

    m_X3SD16->Trigger(X3_SD16::tInputOutput, false);
    m_X3SD16->Trigger(X3_SD16::tInputOutput, true);

    while (m_X3SD16->Pll().Locked() == 0) {
        printf("Waiting PLL Lock\n", 0);
        sleep(1);
    }
    printf( "PLL Locked\n", 0);

    return 1;
}

int main()
{
  CX3SD16 Daq;

  if (Daq.Init() != 1)
    return -1;

  getchar();

  Daq.Close();

  return 1;
}

